/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: mpi-utils.c,v 1.10 2010/07/01 05:50:47 zlb Exp $ */

#ifdef HAVE_CONFIG_H
# include "phg.h"
#else
# include <stdlib.h>
# include <mpi.h>
# define phgAlloc(size)	malloc(size)
# define phgFree(p)	free(p)
# define USE_MPI 1
#endif

#if USE_MPI

#define TAG 314

static int (*isend)(void *, int, MPI_Datatype, int, int, MPI_Comm,
		MPI_Request *) = MPI_Isend;
static int (*irecv)(void *, int, MPI_Datatype, int, int, MPI_Comm,
		MPI_Request *) = MPI_Irecv;

int
phgAlltoallvBegin(
    void *sendbuf, int *sendcnts, int *sdispls, MPI_Datatype sendtype,
    void *recvbuf, int *recvcnts, int *rdispls, MPI_Datatype recvtype,
    MPI_Comm comm, MPI_Request **reqarray, int *nsends, int *nrecvs)
/* Customized Alltoallv function which does not send/recev 0 length blocks.
   This function is derived from intra_Alltoallv() of MPICH 1.2.6 */
{
    int		size, i, j, rcnt, cnt0, dest, rank;
    MPI_Aint	send_extent, recv_extent;
    int		mpi_errno = MPI_SUCCESS;

    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &rank);
    MPI_Type_extent(sendtype, &send_extent);
    MPI_Type_extent(recvtype, &recv_extent);

    *reqarray = phgAlloc(2 * size * sizeof(MPI_Request));

    if (nsends != NULL)
	*nsends = 0;
    if (nrecvs != NULL)
	*nrecvs = 0;

    rcnt = 0;
    for (i = 0; i < size; i++) {
	dest = (rank + i) % size;
	if (recvcnts[dest] <= 0) continue;
	if ((mpi_errno = irecv(
		(void *)((char *)recvbuf+rdispls[dest]*recv_extent),
		recvcnts[dest], recvtype, dest, TAG, comm, &(*reqarray)[rcnt]))
	    != MPI_SUCCESS) break;
	rcnt++;
    }
    cnt0 = rcnt;

    if (mpi_errno == MPI_SUCCESS) {
	for (i = 0; i < size; i++) {
	    dest = (rank + i) % size;
	    if (sendcnts[dest] <= 0) continue;
	    if ((mpi_errno = isend(
		(void *)((char *)sendbuf+sdispls[dest]*send_extent),
		sendcnts[dest], sendtype, dest, TAG, comm, &(*reqarray)[rcnt]))
		!= MPI_SUCCESS) break;
	    rcnt++;
	}
    }

    if (mpi_errno) {
	for (j = 0; j < rcnt; j++)
	    MPI_Cancel(&(*reqarray)[j]);
	phgFree(*reqarray);
	*reqarray = NULL;
	return mpi_errno;
    }

    *nrecvs = cnt0;
    *nsends = rcnt - cnt0;

    return mpi_errno;
}

int
phgAlltoallvEnd(int nsends, int nrecvs, MPI_Request **reqarray)
/* Customized Alltoallv function which does not send/recev 0 length blocks.
   This function is derived from intra_Alltoallv() of MPICH 1.2.6 */
{
    int		j, rcnt;
    int		mpi_errno = MPI_SUCCESS;
    MPI_Status	*starray;

    if (*reqarray == NULL || (rcnt = nsends + nrecvs) <= 0) {
	phgFree(*reqarray);
	*reqarray = NULL;
	return mpi_errno;
    }

    starray = phgAlloc(rcnt * sizeof(MPI_Status));
    mpi_errno = MPI_Waitall(rcnt, *reqarray, starray);
    if (mpi_errno == MPI_ERR_IN_STATUS) {
	for (j = 0; j < rcnt; j++) {
	    if (starray[j].MPI_ERROR != MPI_SUCCESS)
		    mpi_errno = starray[j].MPI_ERROR;
	}
    }
    phgFree(starray);
    phgFree(*reqarray);
    *reqarray = NULL;

    return mpi_errno;
}

int
phgAlltoallv___(
    void *sendbuf, int *sendcnts, int *sdispls, MPI_Datatype sendtype,
    void *recvbuf, int *recvcnts, int *rdispls, MPI_Datatype recvtype,
    MPI_Comm comm, const char *file, int line)
{
    MPI_Request *reqarray;
    int mpi_errno, scnt, rcnt;
    extern void phgInfo(int, const char *, ...);

    mpi_errno = phgAlltoallvBegin(sendbuf, sendcnts, sdispls, sendtype,
			      recvbuf, recvcnts, rdispls, recvtype,
			      comm, &reqarray, &scnt, &rcnt);
#if 0
    phgInfo(-1, "phgAlltoallv: %s:%d, scnt = %d, rcnt = %d, errno = %d\n",
		file, line, scnt, rcnt, mpi_errno);
#endif
    if (mpi_errno != MPI_SUCCESS)
	return mpi_errno;

    return phgAlltoallvEnd(scnt, rcnt, &reqarray);
}

/* Fortran interface */

void
phgalltoallv_(void *sendbuf, int *sendcnts, int *sdispls, int *sendtype,
	      void *recvbuf, int *recvcnts, int *rdispls, int *recvtype,
	      int *comm, int *errcode)
{
    *errcode = phgAlltoallv___(
			sendbuf, sendcnts, sdispls, MPI_Type_f2c(*sendtype),
		 	recvbuf, recvcnts, rdispls, MPI_Type_f2c(*recvtype),
			MPI_Comm_f2c(*comm),
			__FILE__, __LINE__
		);
		 
}

int
phgAlltoallvInit(
    void *sendbuf, int *sendcnts, int *sdispls, MPI_Datatype sendtype,
    void *recvbuf, int *recvcnts, int *rdispls, MPI_Datatype recvtype,
    MPI_Comm comm, MPI_Request **reqarray, int *nsends, int *nrecvs)
{
    int mpi_errno;

    irecv = MPI_Recv_init;
    isend = MPI_Send_init;
    mpi_errno = phgAlltoallvBegin(sendbuf, sendcnts, sdispls, sendtype,
				   recvbuf, recvcnts, rdispls, recvtype,
				   comm, reqarray, nsends, nrecvs);
    irecv = MPI_Irecv;
    isend = MPI_Isend;

    return mpi_errno;
}

#endif	/* USE_MPI */
